/*
 * wxcutils.cc
 * Command utility functions
 *
 */

// For compilers that support precompilation, includes "wx.h".
#include "wx_prec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx.h"
#endif

#include "wx_hash.h"
#include "wx_types.h"
#include "wx_ipc.h"

#include <iostream.h>
#include <ctype.h>
#include <stdlib.h>

#include "wxextend.h"

// Used to add an object plus type to the hash table
class wxTypedObject: public wxObject
{
 public:
  WXTYPE type;
  wxObject *object;
  void *data;
};

// For writing error messages to
ExtensionErrorClass ExtensionError;

// All objects hashed by integer id (in fact the id is the
// pointer; but we have to hash them, or we would not know if
// they are valid)
wxHashTable wxExtensionObjects(wxKEY_INTEGER);

ExtensionErrorClass& ExtensionErrorClass::operator << (char *s)
{
  ExtensionErrorFunction(s);
  return *this;
}

ExtensionErrorClass& ExtensionErrorClass::operator << (char c)
{
  char buf[2];
  buf[0] = c;
  buf[1] = 0;

  ExtensionErrorFunction(buf);
  return *this;
}

ExtensionErrorClass& ExtensionErrorClass::operator << (int i)
{
  char buf[20];
  sprintf(buf, "%d", i);

  ExtensionErrorFunction(buf);
  return *this;
}

ExtensionErrorClass& ExtensionErrorClass::operator << (long i)
{
  char buf[20];
  sprintf(buf, "%ld", i);

  ExtensionErrorFunction(buf);
  return *this;
}

ExtensionErrorClass& ExtensionErrorClass::operator << (double d)
{
  char buf[20];
  sprintf(buf, "%.2f", d);

  ExtensionErrorFunction(buf);
  return *this;
}

ExtensionErrorClass& ExtensionErrorClass::operator << (float f)
{
  char buf[20];
  sprintf(buf, "%.2f", f);

  ExtensionErrorFunction(buf);
  return *this;
}

void wxInitExtension(void)
{
}

void wxCleanupExtension(void)
{
}

/*
 * Storage/retrieval functions.
 */

void wxStoreThingInTable(long item_id, wxObject *object, WXTYPE type)
{
  wxTypedObject *old_obj = (wxTypedObject *)wxExtensionObjects.Get(item_id);
  if (old_obj)
  {
    ExtensionError << "Warning: hash table clash, item = " << item_id;
    ExtensionError << ", old type = " << old_obj->type << ", new type = " << type << "\n";
    return;
  }
  wxTypedObject *obj = new wxTypedObject;
  obj->object = object;
  obj->type = type;
  obj->data = NULL;
  wxExtensionObjects.Put(item_id, obj);
}

void wxDeleteThingInTable(long item_id)
{
  wxTypedObject *obj = (wxTypedObject *)wxExtensionObjects.Delete(item_id);
  if (obj)
    delete obj;
}

wxObject *wxGetThingFromTable(long item_id, WXTYPE *type)
{
  wxTypedObject *obj = (wxTypedObject *)wxExtensionObjects.Get(item_id);
  if (!obj)
    return NULL;

  if (type)
    *type = obj->type;
  return obj->object;
}

wxObject *wxGetTypedObject(long objectId,
                           WXTYPE expectedType)
{
  WXTYPE type;
  wxObject *obj = wxGetThingFromTable(objectId, &type);
  if (!obj)
    return NULL;
  
  if (!wxSubType(type, expectedType))
  {
    char buf[300];
    sprintf(buf, "Type error: %s found where %s expected.\n",
            wxGetTypeName(type), wxGetTypeName(expectedType));
    ExtensionError << buf;
    return NULL;
  }
  return obj;
}

Bool wxAssociateUserData(WXID id, void* dataPtr)
{
    wxTypedObject* obj = (wxTypedObject*)wxExtensionObjects.Get(id);
    if (!obj)
        return FALSE;

    obj->data = dataPtr;
    return TRUE;
}


void* wxGetUserData(WXID id)
{
    wxTypedObject* obj = (wxTypedObject*)wxExtensionObjects.Get(id);
    if (!obj)
        return NULL;

    return obj->data;
}


/*
 * Style flag parsing. Want to be able to pass C-style
 * bitwise-or'ed flags to CLIPS functions, using strings
 * instead of bit lists, e.g.
 * "wxSYSTEM_MENU | wxBORDER"
 */

char *wxExtensionParseWord(char *s, int *i)
{
  if (!s)
    return NULL;

  static char buf[150];
  int len = strlen(s);
  int j = 0;
  int ii = *i;
  while ((ii < len) && (isalpha(s[ii]) || (s[ii] == '_')))
  {
    buf[j] = s[ii];
    j ++;
    ii ++;
  }
  buf[j] = 0;

  // Eat whitespace and conjunction characters
  while ((ii < len) &&
         ((s[ii] == ' ') || (s[ii] == '|') || (s[ii] == ',')))
  {
    ii ++;
  }
  *i = ii;
  if (j == 0)
    return NULL;
  else
    return buf;
}

// Speed-up average search by sorting list
#define compare_func strcmp
static int Compare(const void *elem1, const void *elem2)
{
  const char *str1 = ((wxExtensionBitListStruct *)elem1)->word;
  const char *str2 = ((wxExtensionBitListStruct *)elem2)->word;

  return compare_func(str1, str2);
}

// For a substantialy larger list should use self-adjusting
// trees to optimize lookups
static Bool BitValue (register const char *word, long *val,
 wxExtensionBitListStruct *list, int count)
{
  if (word && *word)
    {
      int High = count;
      int Low = 0;
      int Middle = 0;
      int Old;
      int x;

      do
	{
	  Old = Middle;
	  Middle = (Low + High) / 2;
	  x = compare_func (word, list[Middle].word);
	  if (x < 0)
	    High = Middle;
	  if (x > 0)
	    Low = Middle;
	}
      while (Middle != Old && x != 0);
      if (x == 0) {
	*val = list[Middle].bits; // Found it
	return TRUE;
      }
    }
  return FALSE;
}


wxExtensionBitListStruct wxExtensionBitListTable[] =
{
  { "wxSINGLE", wxSINGLE },
  { "wxMULTIPLE", wxMULTIPLE },
  { "wxEXTENDED", wxEXTENDED },
  { "wxNEEDED_SB", wxNEEDED_SB },
  { "wxALWAYS_SB", wxALWAYS_SB },
  { "wxVSCROLL", wxVSCROLL },
  { "wxHSCROLL", wxHSCROLL },
  { "wxCAPTION", wxCAPTION },
  { "wxABSOLUTE_POSITIONING", wxABSOLUTE_POSITIONING },
  { "wxSTAY_ON_TOP", wxSTAY_ON_TOP},
  { "wxICONIZE", wxICONIZE},
  { "wxMINIMIZE", wxICONIZE},
  { "wxMAXIMIZE", wxMAXIMIZE},
  { "wxSDI", wxSDI},
  { "wxMDI_PARENT", wxMDI_PARENT},
  { "wxMDI_CHILD", wxMDI_CHILD},
  { "wxTHICK_FRAME", wxTHICK_FRAME},
  { "wxRESIZE_BORDER", wxRESIZE_BORDER},
  { "wxSYSTEM_MENU", wxSYSTEM_MENU},
  { "wxMINIMIZE_BOX", wxMINIMIZE_BOX},
  { "wxMAXIMIZE_BOX", wxMAXIMIZE_BOX},
  { "wxRESIZE_BOX", wxRESIZE_BOX},
  { "wxDEFAULT_FRAME", wxDEFAULT_FRAME},
  { "wxDEFAULT_DIALOG_STYLE", wxDEFAULT_DIALOG_STYLE},
  { "wxBORDER", wxBORDER},
  { "wxRETAINED", wxRETAINED},
  { "wxEDITABLE", wxEDITABLE},
  { "wxREADONLY", wxREADONLY},
#ifdef wxPASSWORD
  { "wxPASSWORD", wxPASSWORD},
#endif
  { "wxNATIVE_IMPL", wxNATIVE_IMPL},
  { "wxUSER_COLOURS", wxUSER_COLOURS},
  { "wxBACKINGSTORE", wxBACKINGSTORE},
  { "wxFLAT", wxFLAT},
  { "wxMOTIF_RESIZE", wxMOTIF_RESIZE},
  { "wxFIXED_LENGTH", wxFIXED_LENGTH},
  { "wxPROCESS_ENTER", wxPROCESS_ENTER},

  // Enhanced dialog (not yet implemented in wxExtension)
  { "wxBOTTOM_COMMANDS", wxBOTTOM_COMMANDS},
  { "wxRIGHT_COMMANDS", wxRIGHT_COMMANDS},
  { "wxSTATUS_FOOTER", wxSTATUS_FOOTER},
  { "wxNO_STATUS_FOOTER", wxNO_STATUS_FOOTER},
  { "wxNO_CANCEL_BUTTON", wxNO_CANCEL_BUTTON},
  { "wxCANCEL_BUTTON_FIRST", wxCANCEL_BUTTON_FIRST},
  { "wxCANCEL_BUTTON_SECOND", wxCANCEL_BUTTON_SECOND},
  { "wxCANCEL_BUTTON_LAST", wxCANCEL_BUTTON_LAST},
  { "wxENH_DEFAULT", wxENH_DEFAULT},
  { "wxCOLOURED", wxCOLOURED},
  { "wxTINY_CAPTION_HORIZ", wxTINY_CAPTION_HORIZ},
  { "wxTINY_CAPTION_VERT", wxTINY_CAPTION_VERT},

  // Text font families
  { "wxDEFAULT", wxDEFAULT},
  { "wxDECORATIVE", wxDECORATIVE},
  { "wxROMAN", wxROMAN},
  { "wxSCRIPT", wxSCRIPT},
  { "wxSWISS", wxSWISS},
  { "wxMODERN", wxMODERN},
  { "wxTELETYPE", wxTELETYPE},
  { "wxVARIABLE", wxVARIABLE},
  { "wxFIXED", wxFIXED},
  { "wxNORMAL", wxNORMAL},
  { "wxLIGHT", wxLIGHT},
  { "wxBOLD", wxBOLD},
  { "wxITALIC", wxITALIC},
  { "wxSLANT", wxSLANT},
  { "wxSOLID", wxSOLID},
  { "wxDOT", wxDOT},
  { "wxLONG_DASH", wxLONG_DASH},
  { "wxSHORT_DASH", wxSHORT_DASH},
  { "wxDOT_DASH", wxDOT_DASH},
  { "wxUSER_DASH", wxUSER_DASH},
  { "wxTRANSPARENT", wxTRANSPARENT},
  { "wxSTIPPLE", wxSTIPPLE},
  { "wxBDIAGONAL_HATCH", wxBDIAGONAL_HATCH},
  { "wxCROSSDIAG_HATCH", wxCROSSDIAG_HATCH},
  { "wxFDIAGONAL_HATCH", wxFDIAGONAL_HATCH},
  { "wxCROSS_HATCH", wxCROSS_HATCH},
  { "wxHORIZONTAL_HATCH", wxHORIZONTAL_HATCH},
  { "wxVERTICAL_HATCH", wxVERTICAL_HATCH},
  { "wxJOIN_BEVEL", wxJOIN_BEVEL},
  { "wxJOIN_MITER", wxJOIN_MITER},
  { "wxJOIN_ROUND", wxJOIN_ROUND},
  { "wxCAP_ROUND", wxCAP_ROUND},
  { "wxCAP_PRO<JECTING", wxCAP_PROJECTING},
  { "wxCAP_BUTT", wxCAP_BUTT},

  // Logical ops
  { "wxCLEAR", wxCLEAR},
  { "wxXOR", wxXOR},
  { "wxINVERT", wxINVERT},
  { "wxOR_REVERSE", wxOR_REVERSE},
  { "wxAND_REVERSE", wxAND_REVERSE},
  { "wxCOPY", wxCOPY},
  { "wxAND", wxAND},
  { "wxAND_INVERT", wxAND_INVERT},
  { "wxNO_OP", wxNO_OP},
  { "wxNOR", wxNOR},
  { "wxEQUIV", wxEQUIV},
  { "wxSRC_INVERT", wxSRC_INVERT},
  { "wxOR_INVERT", wxOR_INVERT},
  { "wxNAND", wxNAND},
  { "wxOR", wxOR},
  { "wxSET", wxSET},

  { "wxFLOOD_SURFACE", wxFLOOD_SURFACE},
  { "wxFLOOD_BORDER", wxFLOOD_BORDER},
  { "wxODDEVEN_RULE", wxODDEVEN_RULE},
  { "wxWINDING_RULE", wxWINDING_RULE},
  { "wxHORIZONTAL", wxHORIZONTAL},
  { "wxVERTICAL", wxVERTICAL},
  { "wxBOTH", wxBOTH},
  { "wxCENTER_FRAME", wxCENTER_FRAME},
  { "wxOK", wxOK},
  { "wxYES_NO", wxYES_NO},
  { "wxCANCEL", wxCANCEL},
  { "wxYES", wxYES},
  { "wxNO", wxNO},
  { "wxICON_EXCLAMATION", wxICON_EXCLAMATION},
  { "wxICON_HAND", wxICON_HAND},
  { "wxICON_QUESTION", wxICON_QUESTION},
  { "wxICON_INFORMATION", wxICON_INFORMATION},
  { "wxICON_STOP", wxICON_STOP},
  { "wxICON_ASTERISK", wxICON_ASTERISK},
  { "wxICON_MASK", wxICON_MASK},
  { "wxCENTRE", wxCENTRE},
  { "wxCENTER", wxCENTRE},
  { "wxUSER_COLOURS", wxUSER_COLOURS},
  { "wxVERTICAL_LABEL", wxVERTICAL_LABEL},

  // Bitmap types (not strictly styles)
  { "wxBITMAP_TYPE_BMP", wxBITMAP_TYPE_BMP},
  { "wxBITMAP_TYPE_XPM", wxBITMAP_TYPE_XPM},
  { "wxBITMAP_TYPE_XBM", wxBITMAP_TYPE_XBM},
  { "wxBITMAP_TYPE_RESOURCE", wxBITMAP_TYPE_RESOURCE},
  { "wxBITMAP_TYPE_GIF", wxBITMAP_TYPE_GIF},
  { "wxBITMAP_TYPE_TIF", wxBITMAP_TYPE_TIF},
  { "wxBITMAP_TYPE_ICO", wxBITMAP_TYPE_ICO},
  { "wxBITMAP_TYPE_ICO_RESOURCE", wxBITMAP_TYPE_ICO_RESOURCE},
  { "wxBITMAP_TYPE_CUR", wxBITMAP_TYPE_CUR},
  { "wxBITMAP_TYPE_CUR_RESOURCE", wxBITMAP_TYPE_CUR_RESOURCE},
  { "wxBITMAP_TYPE_XBM_DATA", wxBITMAP_TYPE_XBM_DATA},
  { "wxBITMAP_TYPE_XPM_DATA", wxBITMAP_TYPE_XPM_DATA},
  { "wxBITMAP_TYPE_ANY", wxBITMAP_TYPE_ANY},

  // File dialog related flags
  { "wxOPEN", wxOPEN },
  { "wxSAVE", wxSAVE },
  { "wxOVERWRITE_PROMPT", wxOVERWRITE_PROMPT },
  { "wxHIDE_READONLY", wxHIDE_READONLY },

  // Events
  { "wxTYPE_EVENT", wxTYPE_EVENT },
  { "wxTYPE_MOUSE_EVENT", wxTYPE_MOUSE_EVENT },
  { "wxTYPE_KEY_EVENT", wxTYPE_KEY_EVENT },
  { "wxTYPE_COMMAND_EVENT", wxTYPE_COMMAND_EVENT },
  
  { "wxEVENT_TYPE_BUTTON_COMMAND", wxEVENT_TYPE_BUTTON_COMMAND },
  { "wxEVENT_TYPE_CHECKBOX_COMMAND", wxEVENT_TYPE_CHECKBOX_COMMAND },
  { "wxEVENT_TYPE_CHOICE_COMMAND", wxEVENT_TYPE_CHOICE_COMMAND },
  { "wxEVENT_TYPE_LISTBOX_COMMAND", wxEVENT_TYPE_LISTBOX_COMMAND },
  { "wxEVENT_TYPE_TEXT_COMMAND", wxEVENT_TYPE_TEXT_COMMAND },
  { "wxEVENT_TYPE_TEXT_ENTER_COMMAND", wxEVENT_TYPE_TEXT_ENTER_COMMAND },
  { "wxEVENT_TYPE_MULTITEXT_COMMAND", wxEVENT_TYPE_MULTITEXT_COMMAND },
  { "wxEVENT_TYPE_MENU_COMMAND", wxEVENT_TYPE_MENU_COMMAND },
  { "wxEVENT_TYPE_SLIDER_COMMAND", wxEVENT_TYPE_SLIDER_COMMAND },
  { "wxEVENT_TYPE_RADIOBOX_COMMAND", wxEVENT_TYPE_RADIOBOX_COMMAND },
  { "wxEVENT_TYPE_SET_FOCUS", wxEVENT_TYPE_SET_FOCUS },
  { "wxEVENT_TYPE_KILL_FOCUS", wxEVENT_TYPE_KILL_FOCUS },
 
  { "wxEVENT_TYPE_LEFT_DOWN", wxEVENT_TYPE_LEFT_DOWN  },
  { "wxEVENT_TYPE_LEFT_UP", wxEVENT_TYPE_LEFT_UP },
  { "wxEVENT_TYPE_MIDDLE_DOWN", wxEVENT_TYPE_MIDDLE_DOWN },
  { "wxEVENT_TYPE_RIGHT_DOWN", wxEVENT_TYPE_RIGHT_DOWN },
  { "wxEVENT_TYPE_RIGHT_UP", wxEVENT_TYPE_RIGHT_UP },
  { "wxEVENT_TYPE_MOTION", wxEVENT_TYPE_MOTION },
  { "wxEVENT_TYPE_ENTER_WINDOW", wxEVENT_TYPE_ENTER_WINDOW },
  { "wxEVENT_TYPE_LEAVE_WINDOW", wxEVENT_TYPE_LEAVE_WINDOW },
  { "wxEVENT_TYPE_LEFT_DCLICK", wxEVENT_TYPE_LEFT_DCLICK },
  { "wxEVENT_TYPE_MIDDLE_DCLICK", wxEVENT_TYPE_MIDDLE_DCLICK },
  { "wxEVENT_TYPE_RIGHT_DCLICK", wxEVENT_TYPE_RIGHT_DCLICK },

  // Clipboard formats
  { "wxCF_TEXT", wxCF_TEXT },
  { "wxCF_BITMAP", wxCF_BITMAP },
  { "wxCF_METAFILE", wxCF_METAFILE },
  { "wxCF_DIB", wxCF_DIB },
  { "wxCF_OEMTEXT", wxCF_OEMTEXT },

  // Map Modes
  { "MM_TEXT",        MM_TEXT},
  { "MM_ISOTROPIC",   MM_ISOTROPIC},
  { "MM_ANISOTROPIC", MM_ANISOTROPIC},
  { "MM_LOMETRIC",    MM_LOMETRIC},
  { "MM_HIMETRIC",    MM_HIMETRIC},
  { "MM_TWIPS",       MM_TWIPS},
  { "MM_POINTS",      MM_POINTS},
  { "MM_METRIC",      MM_METRIC},

  // File find flags 
  { "wxFILE", wxFILE },
  { "wxDIR",  wxDIR},

  // VERSION
  { "wxVERSION_NUMBER", wxVERSION_NUMBER}
};

int wxExtensionBitListCount = (sizeof(wxExtensionBitListTable)/sizeof(wxExtensionBitListStruct));

long wxExtensionParseBitList(char *bitListString)
{
  int i = 0;
  char *word;
  long bitList = 0;
  static Bool first_time = TRUE;

  // Sort
  if (first_time)
    {
      // First time? Sort the list w.r.t. compare_func()
      qsort (wxExtensionBitListTable, wxExtensionBitListCount,
	 sizeof(wxExtensionBitListStruct), Compare);
      first_time = FALSE;
    }

  while (word = wxExtensionParseWord(bitListString, &i))
  {
    long bits;

    if (! BitValue(word, &bits, wxExtensionBitListTable, wxExtensionBitListCount))
      ExtensionError << "Unrecognised flag " << word << "\n";
    else
      bitList |= bits;
  }

  return bitList;
}

/*
 * Key code: it's convenient for a language to use a string
 * rather than an integer.
 */

wxExtensionBitListStruct wxExtensionKeyCodeTable[] =
{
 { "WXK_BACK", WXK_BACK },
 { "WXK_TAB", WXK_TAB },
 { "WXK_RETURN", WXK_RETURN },
 { "WXK_ESCAPE", WXK_ESCAPE },
 { "WXK_SPACE", WXK_SPACE },
 { "WXK_DELETE", WXK_DELETE },

 { "WXK_START", WXK_START },
 { "WXK_LBUTTON", WXK_LBUTTON },
 { "WXK_RBUTTON", WXK_RBUTTON },
 { "WXK_CANCEL", WXK_CANCEL },
 { "WXK_MBUTTON", WXK_MBUTTON },
 { "WXK_CLEAR", WXK_CLEAR },
 { "WXK_SHIFT", WXK_SHIFT },
 { "WXK_CONTROL", WXK_CONTROL },
 { "WXK_MENU", WXK_MENU },
 { "WXK_PAUSE", WXK_PAUSE },
 { "WXK_CAPITAL", WXK_CAPITAL },
 { "WXK_PRIOR", WXK_PRIOR },
 { "WXK_NEXT", WXK_NEXT },
 { "WXK_END", WXK_END },
 { "WXK_HOME", WXK_HOME },
 { "WXK_LEFT", WXK_LEFT },
 { "WXK_UP", WXK_UP },
 { "WXK_RIGHT", WXK_RIGHT },
 { "WXK_DOWN", WXK_DOWN },
 { "WXK_SELECT", WXK_SELECT },
 { "WXK_PRINT", WXK_PRINT },
 { "WXK_EXECUTE", WXK_EXECUTE },
 { "WXK_SNAPSHOT", WXK_SNAPSHOT },
 { "WXK_INSERT", WXK_INSERT },
 { "WXK_HELP", WXK_HELP },
 { "WXK_NUMPAD0", WXK_NUMPAD0 },
 { "WXK_NUMPAD1", WXK_NUMPAD1 },
 { "WXK_NUMPAD2", WXK_NUMPAD2 },
 { "WXK_NUMPAD3", WXK_NUMPAD3 },
 { "WXK_NUMPAD4", WXK_NUMPAD4 },
 { "WXK_NUMPAD5", WXK_NUMPAD5 },
 { "WXK_NUMPAD6", WXK_NUMPAD6 },
 { "WXK_NUMPAD7", WXK_NUMPAD7 },
 { "WXK_NUMPAD8", WXK_NUMPAD8 },
 { "WXK_NUMPAD9", WXK_NUMPAD9 },
 { "WXK_MULTIPLY", WXK_MULTIPLY },
 { "WXK_ADD", WXK_ADD },
 { "WXK_SEPARATOR", WXK_SEPARATOR },
 { "WXK_SUBTRACT", WXK_SUBTRACT },
 { "WXK_DECIMAL", WXK_DECIMAL },
 { "WXK_DIVIDE", WXK_DIVIDE },
 { "WXK_F1", WXK_F1 },
 { "WXK_F2", WXK_F2 },
 { "WXK_F3", WXK_F3 },
 { "WXK_F4", WXK_F4 },
 { "WXK_F5", WXK_F5 },
 { "WXK_F6", WXK_F6 },
 { "WXK_F7", WXK_F7 },
 { "WXK_F8", WXK_F8 },
 { "WXK_F9", WXK_F9 },
 { "WXK_F10", WXK_F10 },
 { "WXK_F11", WXK_F11 },
 { "WXK_F12", WXK_F12 },
 { "WXK_F13", WXK_F13 },
 { "WXK_F14", WXK_F14 },
 { "WXK_F15", WXK_F15 },
 { "WXK_F16", WXK_F16 },
 { "WXK_F17", WXK_F17 },
 { "WXK_F18", WXK_F18 },
 { "WXK_F19", WXK_F19 },
 { "WXK_F20", WXK_F20 },
 { "WXK_F21", WXK_F21 },
 { "WXK_F22", WXK_F22 },
 { "WXK_F23", WXK_F23 },
 { "WXK_F24", WXK_F24 },
 { "WXK_NUMLOCK", WXK_NUMLOCK },
 { "WXK_SCROLL", WXK_SCROLL }
};

int wxExtensionKeyCodeCount = (sizeof(wxExtensionKeyCodeTable)/sizeof(wxExtensionBitListStruct));

int wxExtensionKeyCodeStringToInt(char *keyCode)
{
  static Bool first_time = TRUE;
  long val;

  if (first_time)
    {
      // First time? Sort the list w.r.t. compare_func()
      qsort (wxExtensionKeyCodeTable, wxExtensionKeyCodeCount,
	 sizeof(wxExtensionBitListStruct), Compare);
      first_time = FALSE;
    }

  if (!keyCode)
    return -1;

  if (BitValue(keyCode, &val, wxExtensionKeyCodeTable, wxExtensionKeyCodeCount))
    return (int)val;
  return -1; // No Keycode
}

char *wxExtensionKeyCodeIntToString(long keyCode)
{
  for (int j = 0; j < wxExtensionKeyCodeCount; j++)
    if (wxExtensionKeyCodeTable[j].bits == keyCode)
      return wxExtensionKeyCodeTable[j].word;
  return NULL;
}

#undef compare_func

/*
 * Event handling
 */

// Table of event handlers, indexed on event name.
// Each entry is a list of type/function pairs
static wxHashTable wxExtendEventHandlers(wxKEY_STRING);

// Structure for storing event handler
class wxExtendEventHandler: public wxObject
{
 public:
  WXTYPE objectType;
  wxExtensionFunction func;
  wxExtendEventHandler(WXTYPE typ, wxExtensionFunction f)
  {
    objectType = typ; func = f;
  }
};

// Must be called by specific language binding to implement
// callbacks. Register the most specific event handler LAST,
// i.e. register a wxWindow OnPaint handler first, then
// the wxCanvas OnPaint handler.
void wxRegisterExtensionCallback(WXTYPE objectType, char *eventType,
  wxExtensionFunction cFunction)
{
  wxList *list = (wxList *)wxExtendEventHandlers.Get(eventType);
  if (!list)
  {
    list = new wxList;
    wxExtendEventHandlers.Put(eventType, list);
  }
  list->Append(new wxExtendEventHandler(objectType, cFunction));
} 

// Find a callback based on object type and event. Finds the
// most specific handler that's applicable to this event/object
// (last to be registered for this event)
wxExtensionFunction wxFindExtensionCallback(WXTYPE objectType, char *eventType)
{
  wxList *list = (wxList *)wxExtendEventHandlers.Get(eventType);
  if (!list)
    return (wxExtensionFunction)0;

  // Traverse in reverse (from most specific, according to our
  // registration convention, to most general)
  wxNode *node = list->Last();
  while (node)
  {
    wxExtendEventHandler *handler = (wxExtendEventHandler *)node->Data();
    // Check if it's the right type of object for this handler.
    if (wxSubType(objectType, handler->objectType))
      return handler->func;
    node = node->Previous();
  }
  return (wxExtensionFunction)0;
}

